/*!
 * Special Olympics Javascript Game Library
 * Brian Hrebec, 2010
 *
 * Depends on jquery (tested with 1.4.2)
 */

sojsgl = {} // namespace

/*
 * Reimplemeting certain useful functions which aren't 
 * present in old browsers and Internet Exploder
 */

// String.trim()
if (!('trim' in String.prototype)) {  
    String.prototype.trim = function() { 
        return this.replace(/^\s+|\s+$/g,""); 
    };   
};

/* 
 * Array.reduce()
 * Implemented roughly according to ECMA-262-5
 * TODO: make it exact
 */
if (!('reduce' in Array.prototype)) {  
    Array.prototype.reduce = function(fn, initial) {

        if (initial != undefined) {
            var acc = initial;
        } else {
            if (this.length == 0)
                throw TypeError();
            var acc = this[0];
        }

        for (var i = 0; i < this.length; i++) {
            acc = fn(acc, this[i], i, this);
        }
        
        return acc;
    }
};


/* Sound subsystem */

/*! Sound class
 *
 * Represents a single sound file */
sojsgl.Sound = function(url, mimetype) {
    this.url = String(url);
    this.mimetype = String(mimetype);
    this.sndid = -1;
    sojsgl.Sound.uid = 1;
}

sojsgl.Sound.prototype = {
/*! 
 * Plays the sound, previously stopping all other playing sounds 
 */
play: function() {
          sojsgl.Sound.prototype.stopAll();
          this.playNonBlocking();
      },

/*! 
 * Plays the sound, without stopping any other playing sounds 
 *
 * Note that some browser/plugin combos do not produce the expected result:
 *      ie7 (+ wmp) - works
 *      firefox (+ qt) - does not work 
 *          (the last-played sound cuts the others off)
 *      opera (+ qt) - works
 */
playNonBlocking: function() {
              if ($.browser.msie) {
                  $(document.body).append('<object ' + 
                      'classid="clsid:22D6F312-B0F6-11D0-94AB-0080C74C7E95" ' + 
                      'id="sojsgl-snd-' + sojsgl.Sound.uid + '" ' +
                      'class="sojsgl-snd" ' +
                      'width="0" height="0">' +
                      '<param name="FileName" value="' + this.url + '"> ' +
                          '<param name="autoplay" value="true">' + 
                          '<param name="ShowControls" value="false">' + 
                          '<param name="autoStart" value="1"></object>' );
                          
              } else {
                  /* Kludge warning:
                     This is non-standard; future failure will likely be
                     contained here.
                     */
                  $(document.body).append('<embed ' +
                      'id="sojsgl-snd-' + sojsgl.Sound.uid + '" ' +
                      'class="sojsgl-snd" ' + 
                      'name="plugin" ' + 
                      'src="' + this.url + '" ' + 
                      'type="' + this.mimetype + '" ' +
                      'width="0" height="0" autostart="true">' +
                      '<param name="hidden" value="true">'
                      + '</embed>' );
              }
              this.sndid = sojsgl.Sound.uid;
              sojsgl.Sound.uid += 1;
          },
/*! Stops the sound if it is currently playing */
stop: function() {
          if (this.sndid != -1) {
              $('#sojsgl-snd-' + this.sndid).remove();
              this.sndid = -1;
          }
      },
/*! Stops all sounds currently playing */
stopAll: function() {
          $('.sojsgl-snd').each(function(i,e) {
                  $(e).remove();
              });
      }
}

/* end Sound subsystem */

/* Game/Scene subsystem */

/*!
 * Base object for a single game.
 *
 * Each game is a collection of Scenes.
 * From the browser's point of view, the game is 
 * a single html document, allowing game state to
 * be kept across numerous pages from the user's perspective.
 *
 * The Game class also has facilities for organizing and localizing
 * external data used by the game.
 *
 * The expected directory structure is:
 * [pre]
 * [html root]/
 *     [baseDataDir]/
 *          languages.xml
 *          global/
 *              ...data...
 *          [lang_code_1]/
 *              ...data...
 *          [lang_code_2]/
 *              ...data...
 *          [...]/
 *          [lang_code_n]/
 * </pre>
 *
 * Languages.xml is read during the Game creation,
 * and provides a listing of included languages in the following format
 *
 * <pre>
 * language.xml
 * [language code 1],[language_name 1]
 * [language code 2],[language_name 2]
 * [...]
 * </pre>
 * 
 * A browser cookie is kept of the current language setting.
 * The first language in Languages.xml is selected if the 
 * cookie is not present.  
 *
 * To mark an elements as localizable, give it class='tr' and
 * a css id. The id must match the string key (see Game.loadStrings()).
 * When a scene is loaded or the language changed, all so marked elements
 * will be filled (using .html()) with the matching string. (For INPUT elements, 
 * the value attribute will be set instead).
 */
sojsgl.Game = function(name, baseDataDir) {
    
    this.name = name;
    this.m_unload_msg = function () { return ""; };

	this.string_files = {};
    this.strings = {};
    this.userdata = {};
	this.ontranslate = function() {} // TODO: make this cooler

    if (baseDataDir.length > 0 && baseDataDir.charAt(baseDataDir.length - 1) != '/')
        baseDataDir = baseDataDir + '/';

    this.baseDir = baseDataDir;


    var m_currentScene;
    var scenes = {};
    /* adds the given scene to this game */
    this.addScene = function(name, newScene) {
        newScene.game = this;
        scenes[name] = newScene;
    }

    this.currentScene = function() {
        return m_currentScene;
    }

    this.setCurrentScene = function(s) {
        m_currentScene = s;
    }

    this.getScene = function(name) {
        return scenes[name];
    }

    this.setting = function(key) {
        var cookies = document.cookie.split(';');
        key = 'sojsgl_' + this.name + '_' + key + '=';
        for (var i = 0; i < cookies.length; i++) {
            c = cookies[i].trim();
            if (c.indexOf(key) == 0)
                return c.substring(key.length);
        }
        return null;
    }

    this.setSetting = function(key, value) {
        var date = new Date();
		date.setTime(date.getTime()+86400000000); // 1000 days in the future
		var expires = "; expires=" + date.toGMTString();
    
        document.cookie = 'sojsgl_' + this.name + '_' + key + '=' + value + expires + '; path=/';
    }

	this.langs = {};
	that = this;
    // initialize language settings
    this.lang = this.setting('language');
	$.ajax({
		url: this.baseDir + 'languages.xml',
		async: false,
		success: function(data) {
			var xml = sojsgl.parseXML(data);
			$(xml).find('Language').each(function (i, e) {
				var l = $(e);
				code = l.find('Code').text();
				name = l.find('Name').text();

				if (this.lang == null && l.attr('default') == '1')
					this.lang = code;
				that.langs[code] = name;
			});
		}
     });

	if (this.lang == null)
		this.lang = 'en'; //TODO: detect browser language

	this.setSetting('language', this.lang);
}

/*! 
 *  Returns an unattached jquery object containing a language settings
 *  form following this template:
 *  <pre>
 *  &lt;form&gt;&lt;ul&gt;
 *  &lt;li&gt;
 *      &lt;input type='radio' name='language' value='[language code 1]' &gt;
 *      [language name 1]
 *  &lt;/li&gt;
 *  &lt;li&gt;
 *      &lt;input type='radio' name='language' value='[language code 2]' &gt;
 *      [language name 2]
 *  &lt;/li&gt;
 *  [...]
 *
 *  &lt;li&gt;
 *      &lt;input type='radio' name='language' value='[language code n]' &gt;
 *      [language name n]
 *  &lt;/li&gt;
 *  &lt;/ul&gt;&lt;/form&gt;
 *  </pre>
 *
 *  When the user selects a language, the game's language will be immediately changed.
 */
sojsgl.Game.prototype.getLanguageMenu = function() {
    var that = this;
	var acc = '';
	$.each(this.langs, function(k, v) {
		acc = acc + '<li><input type="radio" name="language" value="' + 
				k + '" ' + (that.lang == k ? 'checked' : '') + ' >' + v + '</li>';
	});
    return $('<form />').html('<ul>' + acc + '</ul>')
                        .click(function(e) {
                                if (e.target.tagName == 'INPUT' && 
                                        e.target.type == 'radio')
                                    that.setLanguage(e.target.value);
                            }
                        );
}

/* 
 * When called without an argument, translates all the items
 * currently present. 
 * 
 * When called with an argument, also sets the game's current language
 * and saves it to the language cookie.
 *
 * \param lang language to set the game to
 */
sojsgl.Game.prototype.setLanguage = function(lang) {
    var that = this;
	var old_lang = this.lang;
    if (lang != undefined && lang != old_lang) {
		this.lang = lang;
		this.setSetting('language', this.lang);
		$.each(this.string_files, function(k, v) {
					if (v == old_lang) {
						that.loadStrings(k);
					}
				});
	}

    $('.tr').each(function() { 
            if (this.tagName == 'INPUT')
                $(this).val(that.strings[this.id]);
            else
                $(this).html(that.strings[this.id]);
        });

	this.ontranslate();
}

/*!
 * enables or disables confirmation when the user attempts to 
 * navigate away from the game.
 *
 *  unload - if true, enable
 */
sojsgl.Game.prototype.setConfirmExit = function(unload) {
    if (unload)
        window.onbeforeunload = this.m_unload_msg;
    else
        window.onbeforeunload = null;
}

/* Set the message used for exit confirmation.
   Does not affect whether confirmation is enabled.
 */
sojsgl.Game.prototype.setConfirmMessage = function(msg) {
    this.m_unload_msg = function() {
        return msg;
    }

    if (window.onbeforeunload != null)
        this.setConfirmExit(true);
}


//! Loads string resources from the given file.
/*!
 * A string file is an XML file of the form:
 * <Strings>
 *   <String>
 *	   <Key>StringKey</Key>
 *	   <Value>StringVlaue</Value>
 *   </String>
 * </Strings>
 *   
 *
 * \param stringFile The path, relative to the game's base dir, 
 *                   to the desired string file
 * \param language If defined, overrides the language dir setting. 
 *                 (Especially useful for global resources)
 */
sojsgl.Game.prototype.loadStrings = function(stringFile, language) {
    var that = this;
	language = language == undefined ? this.lang : language;
	that.string_files[stringFile] = language;
	$.ajax({url: this.baseDir + language + '/' + stringFile, 
		   success: function(data) {
			  xml = $(sojsgl.parseXML(data));
			  xml.find('String').each(function(i, e) {
			  	  that.strings[$(e).find('Key').text()] = $(e).find('Value').text();
			  });
		   },
		   async: false
	   });
}


//! Loads object resources from the given file.
/*!
 * Object files are XML files of the following form
 * <SOJSGLObjects>
 *   <Object>
 *     <Prop1>Prop1Value</Prop1>
 *     <Prop2>Prop1Value</Prop2>
 *     ....
 *     <PropN>Prop1Value</PropN>
 *   </Object>
 * </SOJSGLObjects>
 *
 * Where Prop1 ... PropN are arbitrary object property names
 *
 * \param constructor A function, which will be called precisely once for each defined
 *                    object. It will be passed as its sole argument a map of 
 *                    property names to propery values.
 * \param objectFile The path, relative to the game's base dir, to the desired file
 * \param language If defined, overrides the language dir setting. 
 *                 (Especially useful for global resources)
 */
sojsgl.Game.prototype.loadObjects = function(constructor, objectFile, language) {
	language = language == undefined ? this.lang : language;
	$.ajax({url: this.baseDir + language + '/' + objectFile, 
			success: function(data) {
				xml = $(sojsgl.parseXML(data));
				xml.find('Object').each(function (i, e) {
                    props = {};
                    $(e).children().each(function(i, e) {
						props[e.tagName] = e.textContent;
                    });
                    constructor(props);
				});
			},
			async: false
	});
}


//! Scene class. 
/*! 
 * Each scene represents some arbitrary stage of the game.
 * This allow a game to consist of multiple 'pages' without requiring a
 * server to store game state data. The scene class dynamically loads the
 * pages into the same html document.
 *
 * Each scene is contained in an external html file, the form of which 
 * is arbitrary.
 *
 * Within that file, use
 *      sojsgl.Game.currentGame.currentScene()
 * to obtain a reference to the current Scene object, to (for instance) add 
 * event handlers, preEnd effects, etc. This will allow multiple instances of the
 * same scene to be created and for scenes to shared between games.
 */

/* Scene constructor

   htmldoc - uri to the scene file
   container - jquery selector to the DOM object to display this scene in.
               Note that the contents of 'container' will be erased.
 */
sojsgl.Scene = function(htmldoc, container) {
    this.m_htmldoc = htmldoc;
    this.m_container = container;
    this.m_isloaded = false;
    this.sprites = {};
    this.m_postLoad = [];
    this.m_preEnd = [];

    this.nextScene = this;
}

sojsgl.Scene.prototype = {

/* loads this scene and displays it */
begin: function() {
        if (this.m_isloaded || this.game == undefined) {
            return;
        }
        sojsgl.Game.currentGame = this.game;
        this.game.setCurrentScene(this);
        var that = this;
        if (this.m_htmldoc) {
            $.ajax({url: this.m_htmldoc, 
                    success: function(data) {
							$(that.m_container).html(data);

                            m_isloaded = true;
                            for (s in that.sprites) {
                                s.paint();
                            }
                            for (f in that.m_postBegin) {
                                f.func();
                            }

							that.game.setLanguage(); // localize the scene
                        },
                    beforeSend: function(xhr) { 
                            if (xhr.overrideMimeType) xhr.overrideMimeType("text/plain"); 
                        }
                    });
        }

    },
/* Hides this scene and performs associated cleanup.

   nextscene - if defined, nextscene.begin() will be called 
               upon this scene's end
   removal_function - if defined, will be used to remove the Scene 
                      from the DOM instead of the default removal function
 */
end: function(nextscene, removal_function) {
        var t = 0;
        var that = this;
        $.each (this.m_preEnd, function() {
            if (this.time > t) {
                t = this.time;
            }
            this.func();
        });

        setTimeout(function() {
            if (removal_function) {
                removal_function();
            } else {
                $(that.m_container).empty(); 
            }

            for (p in that.sprites) {
                p.hide();
            }
            this.m_isloaded = false;
            if (nextscene) {
                nextscene.begin();
            }
        }, t);
    },
/* Adds a sprite to the Scene for display, drag&drop, etc.
 */
addSprite: function(sprite) {
        sprites[sprite.name()] = sprite;
    },
/* Removes a sprite from the Scene
 */
removeSprite: function(sprite) {
        delete sprites[sprite.name()];
    },
/* adds a function to be executed immedately after the Scene is loaded:

   f - function to execute
   time - amount of time estimated for this function (for, eg. visual effects)
   */
addPostBegin: function(f, time) {
        this.m_postBegin.push({func: f, time: time});
    },
/* adds a function to be executed immedately before the Scene is unloaded:

   f - function to execute
   time - amount of time estimated for this function (for, eg. visual effects)
          removal of this scene from the DOM will be delayed by this amount.
   */
addPreEnd: function(f, time) {
        this.m_preEnd.push({func: f, time: time});
    }
}

/* Dialog class. A Dialog is a special Scene implementing a 
   general modal dialog of the form:

        <div class="sojsgl-dialog-wrapper">
            <div class="sojsgl-dialog-alpha-wrapper" />
            <div class="sojsgl-dialog" />
				<div class="sojsgl-dialog-inner" />
					[contents]
					[buttons - <input type="button" class="sojsgl-dialog-button"]
				</div>
			</div>
        </div>

   By default, the dialog will contain a single 'OK' button. Other buttons
   can be specified by using setButtons().

   contents - html to be placed in the dialog 
   container - jquery selector for element to place dialog inside
 */
sojsgl.Dialog = function(contents, container) {
    this.m_contents = contents;
    this.m_container = container;
    var that = this;
    this.m_buttons = { 'OK': function() { that.end(); } };
}

sojsgl.Dialog.prototype = new sojsgl.Scene();
sojsgl.Dialog.prototype.constructor = sojsgl.Dialog;

/* Redefines the buttons for this dialog 

    buttons - button definitions in the form:
            { [button_name]: [button_function()], ... } 
 */
sojsgl.Dialog.prototype.setButtons = function(buttons) {
    this.m_buttons = buttons;
}

sojsgl.Dialog.prototype.begin = function() {
    //TODO: get rid of this conditional
    if (this.m_isloaded || this.game == undefined) {
        return;
    }

    sojsgl.Scene.prototype.begin.call(this);
    this.uid = 'dia-' + (new Date).getTime();
    var dia = $('<div />').attr('class', 'sojsgl-dialog').
		append($('<div />').addClass('sojsgl-dialog-inner').
        append(this.m_contents));
    var p = dia.append($('<p />'));
    $.each(this.m_buttons, function(k, v) {
            p.append($('<input type="button" />').attr({
                    'class' : 'sojsgl-dialog-button',
                   'value' : k}).click(v));
        }); 
    $(this.m_container).append(
            $('<div />').attr({
                'id': this.uid,
                'class': 'sojsgl-dialog-wrapper' }).
                append($('<div class="sojsgl-dialog-alpha-wrapper" />')).
                append(dia));
}

sojsgl.Dialog.prototype.end = function() {
    var that = this;
    sojsgl.Scene.prototype.end.call(this, undefined, function() {
            $('#' + that.uid).remove();
        }); // handle removal manually
}

/* end Game/Scene subsystem */

/* Graphics subsystem */

/*! 
 * Sprite constructor:
 * Creates a sprite.
 *
 * image - uri for the sprite's inital image
 * parent - jquery selector for the object to place the sprite in
 * width - initial width in pixels
 * height - initial height in pixels
 */
sojsgl.Sprite = function(image, parent, width, height) {
    this.x = 0; 
    this.y = 0;
    this.w = width;
    this.h = height;
    this.m_parent = parent;
    this.animmode = 0;
    this.frames = [];

    this.init = function() {
        this.m_self = $('<img class="sojsgl-sprite" src="' + 
                this.frames[0] + '" />').
            css({'position': 'absolute', 'top': this.y, 'left': this.x,
                    'width': this.w, 'height': this.h, 'display': 'none' });

        $(this.m_parent).append(this.m_self);
    };

    if (image) {
        this.frames.push(image);
        this.init();
    }

    this.current_frame = 0;
    this.timer_id = -1;


}

sojsgl.Sprite.prototype.jquery_ref = function() { return this.m_self; }

/*
sojsgl.Sprite.Origins = { 'topleft':      [0,   0], 
                          'topcenter':    [0.5, 0], 
                          'topright':     [1,   0],
                          'centerleft':   [0,   0.5],  
                          'center':       [0.5, 0.5], 
                          'centerright':  [1,   0.5],
                          'bottomleft':   [0,   1], 
                          'bottomcenter': [0.5, 1], 
                          'bottomright':  [1,   1]
                        }
                        */

sojsgl.Sprite.prototype.renderAt = function(lx, ly, lw, lh) {
    if (lw) this.w = lw;
    if (lh) this.h = lh;
    this.x = lx;
    this.y = ly;
    this.m_self.css({'top': ly, 'left': lx, 'width': lw, 'height': lh }).show();

}

/*!
 * Moves the sprite to a new location and resizes it, all over the given time.
 *
 * t - time of animation, defaults to 0 (instant)
 * lx - new x, defaults to no change
 * ly - new y, defaults to no change
 * lw - new width, defaults to no change
 * lh - new height, defaults to no change
 */
sojsgl.Sprite.prototype.moveTo = function(t, lx, ly, lw, lh, easing) {
    if (t == undefined) t = 0;
    if (lx == undefined) lx = this.x;
    if (ly == undefined) ly = this.y;
    if (lw == undefined) lw = this.w;
    if (lh == undefined) lh = this.h;
    if (easing == undefined) easing = 'swing';

    this.x = lx;
    this.y = ly;
    this.w = lw;
    this.h = lh;

    this.m_self.animate({'top': this.y, 'left': this.x, 
            'width': this.w, 'height': this.h },
            { 'duration': t, 'easing': easing }
            );
}

sojsgl.Sprite.prototype.setVisible = function(vis) {
    if (vis)
        this.m_self.show();
    else
        this.m_self.hide();
}

sojsgl.Sprite.prototype.addFrames = function(images) {
    var was_null = false;
    if (frames.length == 0)
        was_null = true;

    if ($.isArray(images)) {
        for (i in images) {
            this.frames.push(i);
        }
    } else {
        this.frames.push(images);
    }

    if (was_null)
        this.init();
}

sojsgl.Sprite.prototype.addFramesFrom = function(path, prefix, count, ext) {
    if (path.charAt(path.length - 1) != '/')
        path = path + '/';

    if (ext.charAt(0) != '.')
        ext = '.' + ext;

    for (i = 0; i < count; i++) {
        this.addFrames(path + prefix + i + ext);
    }
}


/* */
sojsgl.Sprite.prototype.animate = function(rate, mode, repeats) {
    if (frames.length < 2)
        return;
    var that = this;
    
    if (mode == 'pingpong') {
        var direction = 1;
        this.timer_id = setInterval(function() {
                if (that.current_frame == that.frames.length - 1 ||
                        that.current_frame == 0) {
                    if (repeats == 0) {
                        clearInterval(that.timer_id);
                        return;
                    } else {
                        repeats--;
                        direction = -direction;
                    }
                }
                that.current_frame += direction;

                that.m_self.css('src', that.frames[current_frame]);
            });
    } else { //circular
        this.timer_id = setInterval(function() {
                if (that.current_frame == that.frames.length - 1) {
                    if (repeats == 0) {
                        clearInterval(that.timer_id);
                        return;
                    } else {
                        repeats--;
                        that.current_frame = 0;
                    }
                } else {
                    that.current_frame++;
                }

                that.m_self.css('src', that.frames[current_frame]);
            });
    }
}

sojsgl.Sprite.prototype.animateStop = function() {
    clearInterval(this.timer_id);
}

sojsgl.Sprite.prototype.destroy = function() {
    this.m_self.remove();
}

/* end Graphics subsystem */

/* Utility functions */

/* Just in case the server mime type is screwed up (esp. for local files */
sojsgl.parseXML = function(data) {
	var xml;
	if (typeof data == "string") {
		if( window.DOMParser ) {
			parser = new DOMParser();
			xml = parser.parseFromString(data, "text/xml") ;
		} else { // internet exploder...
		   xml = new ActiveXObject("Microsoft.XMLDOM");
		   xml.async = false;
		   xml.loadXML(data);
		}
	} else { // assume it's an XML object
		xml = data;
	}
	return xml;
}

sojsgl.strParamReplace = function(str, map) {
    for (k in map) {
        // TODO: test this
        // TODO: deal with special characters in k
        var result = str.replace(/\$/g, '$$$$');
        str = str.replace(RegExp('{' + k + '}', 'g'), map[k]);
    }

    return str;
}
/* end Utility functions */
